"""This module contains RQ jobs"""
from rq.decorators import job
from rq import get_current_job
from xcessiv import functions
from xcessiv import exceptions
from xcessiv import models
import numpy as np
import os
import sys
import traceback
from sklearn.model_selection import train_test_split


def extraction_data_statistics(path):
    """ Generates data statistics for the given data extraction setup stored
    in Xcessiv notebook.

    This is in rqtasks.py but not as a job yet. Temporarily call this directly
    while I'm figuring out Javascript lel.

    Args:
        path (str, unicode): Path to xcessiv notebook
    """
    with functions.DBContextManager(path) as session:
        extraction = session.query(models.Extraction).first()
        X, y = extraction.return_main_dataset()
        functions.verify_dataset(X, y)

        if extraction.test_dataset['method'] == 'split_from_main':
            X, X_test, y, y_test = train_test_split(
                X,
                y,
                test_size=extraction.test_dataset['split_ratio'],
                random_state=extraction.test_dataset['split_seed'],
                stratify=y
            )
        elif extraction.test_dataset['method'] == 'source':
            if 'source' not in extraction.test_dataset or not extraction.test_dataset['source']:
                raise exceptions.UserError('Source is empty')

            extraction_code = extraction.test_dataset["source"]
            extraction_function = functions.\
                import_object_from_string_code(extraction_code, "extract_test_dataset")
            X_test, y_test = extraction_function()
        else:
            X_test, y_test = None, None

        # test base learner cross-validation
        extraction_code = extraction.meta_feature_generation['source']
        return_splits_iterable = functions.import_object_from_string_code(
            extraction_code,
            'return_splits_iterable'
        )
        number_of_splits = 0
        test_indices = []
        try:
            for train_idx, test_idx in return_splits_iterable(X, y):
                number_of_splits += 1
                test_indices.append(test_idx)
        except Exception as e:
            raise exceptions.UserError('User code exception', exception_message=str(e))

        # preparation before testing stacked ensemble cross-validation
        test_indices = np.concatenate(test_indices)
        X, y = X[test_indices], y[test_indices]

        # test stacked ensemble cross-validation
        extraction_code = extraction.stacked_ensemble_cv['source']
        return_splits_iterable = functions.import_object_from_string_code(
            extraction_code,
            'return_splits_iterable'
        )
        number_of_splits_stacked_cv = 0
        try:
            for train_idx, test_idx in return_splits_iterable(X, y):
                number_of_splits_stacked_cv += 1
        except Exception as e:
            raise exceptions.UserError('User code exception', exception_message=str(e))

        data_stats = dict()
        data_stats['train_data_stats'] = functions.verify_dataset(X, y)
        if X_test is not None:
            data_stats['test_data_stats'] = functions.verify_dataset(X_test, y_test)
        else:
            data_stats['test_data_stats'] = None
        data_stats['holdout_data_stats'] = {'number_of_splits': number_of_splits}
        data_stats['stacked_ensemble_cv_stats'] = {'number_of_splits': number_of_splits_stacked_cv}

        extraction.data_statistics = data_stats

        session.add(extraction)
        session.commit()


@job('default', timeout=86400)
def generate_meta_features(path, base_learner_id):
    """Generates meta-features for specified base learner

    After generation of meta-features, the file is saved into the meta-features folder

    Args:
        path (str): Path to Xcessiv notebook

        base_learner_id (str): Base learner ID
    """
    with functions.DBContextManager(path) as session:
        base_learner = session.query(models.BaseLearner).filter_by(id=base_learner_id).first()
        if not base_learner:
            raise exceptions.UserError('Base learner {} '
                                       'does not exist'.format(base_learner_id))

        base_learner.job_id = get_current_job().id
        base_learner.job_status = 'started'

        session.add(base_learner)
        session.commit()

        try:
            est = base_learner.return_estimator()
            extraction = session.query(models.Extraction).first()
            X, y = extraction.return_train_dataset()
            return_splits_iterable = functions.import_object_from_string_code(
                extraction.meta_feature_generation['source'],
                'return_splits_iterable'
            )

            meta_features_list = []
            trues_list = []
            for train_index, test_index in return_splits_iterable(X, y):
                X_train, X_test = X[train_index], X[test_index]
                y_train, y_test = y[train_index], y[test_index]
                est = est.fit(X_train, y_train)
                meta_features_list.append(
                    getattr(est, base_learner.base_learner_origin.
                            meta_feature_generator)(X_test)
                )
                trues_list.append(y_test)
            meta_features = np.concatenate(meta_features_list, axis=0)
            y_true = np.concatenate(trues_list)

            for key in base_learner.base_learner_origin.metric_generators:
                metric_generator = functions.import_object_from_string_code(
                    base_learner.base_learner_origin.metric_generators[key],
                    'metric_generator'
                )
                base_learner.individual_score[key] = metric_generator(y_true, meta_features)

            meta_features_path = base_learner.meta_features_path(path)

            if not os.path.exists(os.path.dirname(meta_features_path)):
                os.makedirs(os.path.dirname(meta_features_path))

            np.save(meta_features_path, meta_features, allow_pickle=False)
            base_learner.job_status = 'finished'
            base_learner.meta_features_exists = True
            session.add(base_learner)
            session.commit()

        except:
            session.rollback()
            base_learner.job_status = 'errored'
            base_learner.description['error_type'] = repr(sys.exc_info()[0])
            base_learner.description['error_value'] = repr(sys.exc_info()[1])
            base_learner.description['error_traceback'] = \
                traceback.format_exception(*sys.exc_info())
            session.add(base_learner)
            session.commit()
            raise


@job('default', timeout=86400)
def evaluate_stacked_ensemble(path, ensemble_id):
    """Evaluates the ensemble and updates the database when finished/

    Args:
        path (str): Path to Xcessiv notebook

        ensemble_id (str): Ensemble ID
    """
    with functions.DBContextManager(path) as session:
        stacked_ensemble = session.query(models.StackedEnsemble).filter_by(
            id=ensemble_id).first()
        if not stacked_ensemble:
            raise exceptions.UserError('Stacked ensemble {} '
                                       'does not exist'.format(ensemble_id))

        stacked_ensemble.job_id = get_current_job().id
        stacked_ensemble.job_status = 'started'

        session.add(stacked_ensemble)
        session.commit()

        try:
            meta_features_list = []
            for base_learner in stacked_ensemble.base_learners:
                mf = np.load(base_learner.meta_features_path(path))
                if len(mf.shape) == 1:
                    mf = mf.reshape(-1, 1)
                meta_features_list.append(mf)

            secondary_features = np.concatenate(meta_features_list, axis=1)

            # Get data
            extraction = session.query(models.Extraction).first()
            return_splits_iterable = functions.import_object_from_string_code(
                extraction.meta_feature_generation['source'],
                'return_splits_iterable'
            )
            X, y = extraction.return_train_dataset()

            #  We need to retrieve original order of meta-features
            indices_list = [test_index for train_index, test_index in return_splits_iterable(X, y)]
            indices = np.concatenate(indices_list)
            X, y = X[indices], y[indices]

            if stacked_ensemble.append_original:
                secondary_features = np.concatenate((secondary_features, X), axis=1)

            est = stacked_ensemble.return_secondary_learner()

            return_splits_iterable_stacked_ensemble = functions.import_object_from_string_code(
                extraction.stacked_ensemble_cv['source'],
                'return_splits_iterable'
            )
            preds = []
            trues_list = []
            for train_index, test_index in return_splits_iterable_stacked_ensemble(secondary_features, y):
                X_train, X_test = secondary_features[train_index], secondary_features[test_index]
                y_train, y_test = y[train_index], y[test_index]
                est = est.fit(X_train, y_train)
                preds.append(
                    getattr(est, stacked_ensemble.base_learner_origin.
                            meta_feature_generator)(X_test)
                )
                trues_list.append(y_test)
            preds = np.concatenate(preds, axis=0)
            y_true = np.concatenate(trues_list)

            for key in stacked_ensemble.base_learner_origin.metric_generators:
                metric_generator = functions.import_object_from_string_code(
                    stacked_ensemble.base_learner_origin.metric_generators[key],
                    'metric_generator'
                )
                stacked_ensemble.individual_score[key] = metric_generator(y_true, preds)

            stacked_ensemble.job_status = 'finished'
            session.add(stacked_ensemble)
            session.commit()

        except:
            session.rollback()
            stacked_ensemble.job_status = 'errored'
            stacked_ensemble.description['error_type'] = repr(sys.exc_info()[0])
            stacked_ensemble.description['error_value'] = repr(sys.exc_info()[1])
            stacked_ensemble.description['error_traceback'] = \
                traceback.format_exception(*sys.exc_info())
            session.add(stacked_ensemble)
            session.commit()
            raise



from __future__ import absolute_import, print_function, division, unicode_literals
from xcessiv import app
import sys


if __name__ == '__main__':
    port = 8080 if len(sys.argv) < 2 else sys.argv[1]
    app.run(debug=True, port=int(port), host='0.0.0.0')



from __future__ import absolute_import, print_function, division, unicode_literals
from gevent.wsgi import WSGIServer
import webbrowser


def launch(app):
    http_server = WSGIServer(('', app.config['XCESSIV_PORT']), app)
    webbrowser.open_new('http://localhost:' + str(app.config['XCESSIV_PORT']))
    http_server.serve_forever()



from xcessiv.presets.cvsetting import k_fold


extraction_default_source = """\"\"\"In this code block, you must define the function `extract_{}_dataset`.
`extract_{}_dataset` must take no arguments and return a tuple (X, y), where X is a
Numpy array with shape (n_samples, n_features) corresponding to the features of your
{} dataset and y is the Numpy array corresponding to the ground truth labels of each
sample.
\"\"\"

def extract_{}_dataset():
    return [[1, 2], [2, 1]], [0, 1]
"""

meta_feature_generation_default_source = k_fold['source']

DEFAULT_EXTRACTION_MAIN_DATASET = {
    "source": extraction_default_source.format('main', 'main', 'main', 'main')
}
DEFAULT_EXTRACTION_TEST_DATASET = {
    "method": None,
    "source": extraction_default_source.format('test', 'test', 'test', 'test')
}
DEFAULT_EXTRACTION_META_FEATURE_GENERATION = {
    "source": meta_feature_generation_default_source
}



"""This module contains the SQLAlchemy ORM Models"""
from __future__ import absolute_import, print_function, division, unicode_literals
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Text, Integer, Boolean, TypeDecorator, ForeignKey, Table
from sqlalchemy.orm import relationship
from sqlalchemy.ext import mutable
import numpy as np
import json
import os
from sklearn.model_selection import train_test_split
from xcessiv import constants
from xcessiv import exceptions
from xcessiv import functions
from xcessiv import app


Base = declarative_base()


class JsonEncodedDict(TypeDecorator):
    """Enables JSON storage by encoding and decoding on the fly."""
    impl = Text

    def process_bind_param(self, value, dialect):
        return json.dumps(value, sort_keys=True)

    def process_result_value(self, value, dialect):
        return json.loads(value)


mutable.MutableDict.associate_with(JsonEncodedDict)


class Extraction(Base):
    """This table's columns are text columns representing JSON data of how
    to extract the train and test datasets, base learner cross-validation method,
    and the stacked ensemble cross-validation method.
    It will contain only a single row.
    """
    __tablename__ = 'extraction'

    id = Column(Integer, primary_key=True)
    main_dataset = Column(JsonEncodedDict)
    test_dataset = Column(JsonEncodedDict)
    meta_feature_generation = Column(JsonEncodedDict)
    stacked_ensemble_cv = Column(JsonEncodedDict)
    data_statistics = Column(JsonEncodedDict)

    def __init__(self):
        self.main_dataset = constants.DEFAULT_EXTRACTION_MAIN_DATASET
        self.test_dataset = constants.DEFAULT_EXTRACTION_TEST_DATASET
        self.meta_feature_generation = constants.DEFAULT_EXTRACTION_META_FEATURE_GENERATION
        self.stacked_ensemble_cv = constants.DEFAULT_EXTRACTION_META_FEATURE_GENERATION
        self.data_statistics = None

    def return_main_dataset(self):
        """Returns main data set from self

        Returns:
            X (numpy.ndarray): Features

            y (numpy.ndarray): Labels
        """
        if not self.main_dataset['source']:
            raise exceptions.UserError('Source is empty')

        extraction_code = self.main_dataset["source"]

        extraction_function = functions.import_object_from_string_code(extraction_code,
                                                                       "extract_main_dataset")
        try:
            X, y = extraction_function()
        except Exception as e:
            raise exceptions.UserError('User code exception', exception_message=str(e))

        X, y = np.array(X), np.array(y)

        return X, y

    def return_train_dataset(self):
        """Returns train data set

        Returns:
            X (numpy.ndarray): Features

            y (numpy.ndarray): Labels
        """
        X, y = self.return_main_dataset()

        if self.test_dataset['method'] == 'split_from_main':
            X, X_test, y, y_test = train_test_split(
                X,
                y,
                test_size=self.test_dataset['split_ratio'],
                random_state=self.test_dataset['split_seed'],
                stratify=y
            )

        return X, y

    def return_test_dataset(self):
        """Returns test data set

        Returns:
            X (numpy.ndarray): Features

            y (numpy.ndarray): Labels
        """
        if self.test_dataset['method'] == 'split_from_main':
            X, y = self.return_main_dataset()
            X, X_test, y, y_test = train_test_split(
                X,
                y,
                test_size=self.test_dataset['split_ratio'],
                random_state=self.test_dataset['split_seed'],
                stratify=y
            )

            return X_test, y_test

        if self.test_dataset['method'] == 'source':
            if 'source' not in self.test_dataset or not self.test_dataset['source']:
                raise exceptions.UserError('Source is empty')

            extraction_code = self.test_dataset["source"]
            extraction_function = functions.\
                import_object_from_string_code(extraction_code, "extract_test_dataset")
            X_test, y_test = extraction_function()

            return np.array(X_test), np.array(y_test)


class BaseLearnerOrigin(Base):
    """This table contains the base learner origins of the Xcessiv notebook"""
    __tablename__ = 'baselearnerorigin'

    id = Column(Integer, primary_key=True)
    source = Column(Text)
    validation_results = Column(JsonEncodedDict)
    hyperparameters = Column(JsonEncodedDict)
    name = Column(Text)
    final = Column(Boolean)
    meta_feature_generator = Column(Text)
    metric_generators = Column(JsonEncodedDict)
    description = Column(JsonEncodedDict)
    base_learners = relationship('BaseLearner', back_populates='base_learner_origin',
                                 cascade='all, delete-orphan', single_parent=True)
    stacked_ensembles = relationship('StackedEnsemble', back_populates='base_learner_origin',
                                     cascade='all, delete-orphan', single_parent=True)

    def __init__(self, source='', name='Base Learner Setup',
                 meta_feature_generator='predict_proba', metric_generators=None):
        self.source = source
        self.name = name
        self.validation_results = dict()
        self.hyperparameters = dict()
        self.final = False
        self.meta_feature_generator = meta_feature_generator
        self.description = dict()
        self.metric_generators = dict() if metric_generators is None else metric_generators

    @property
    def serialize(self):
        return dict(
            id=self.id,
            source=self.source,
            name=self.name,
            validation_results=self.validation_results,
            final=self.final,
            meta_feature_generator=self.meta_feature_generator,
            metric_generators=self.metric_generators,
            hyperparameters=self.hyperparameters
        )

    def return_estimator(self):
        """Returns estimator from base learner origin

        Returns:
            est (estimator): Estimator object
        """
        extraction_code = self.source
        estimator = functions.import_object_from_string_code(extraction_code, "base_learner")

        return estimator

    def cleanup(self, path):
        """This function should be called before database deletion to do any pre-delete work

        Args:
            path (str, unicode): Absolute/local path of xcessiv folder
        """
        for learner in self.base_learners:
            learner.cleanup(path)


association_table = Table(
    'association', Base.metadata,
    Column('baselearner_id', Integer, ForeignKey('baselearner.id')),
    Column('stackedensemble_id', Integer, ForeignKey('stackedensemble.id'))
)


class BaseLearner(Base):
    """This table contains base learners of the Xcessiv notebook"""
    __tablename__ = 'baselearner'

    id = Column(Integer, primary_key=True)
    hyperparameters = Column(JsonEncodedDict)
    individual_score = Column(JsonEncodedDict)
    meta_features_exists = Column(Boolean)
    job_status = Column(Text)
    job_id = Column(Text)
    description = Column(JsonEncodedDict)
    base_learner_origin_id = Column(Integer, ForeignKey('baselearnerorigin.id'))
    base_learner_origin = relationship('BaseLearnerOrigin', back_populates='base_learners')
    stacked_ensembles = relationship(
        'StackedEnsemble',
        secondary=association_table,
        back_populates='base_learners',
        cascade='all, delete-orphan',
        single_parent=True
    )

    def __init__(self, hyperparameters, job_status, base_learner_origin):
        self.hyperparameters = hyperparameters
        self.individual_score = dict()
        self.meta_features_exists = False
        self.job_status = job_status
        self.job_id = None
        self.description = dict()
        self.base_learner_origin = base_learner_origin

    def return_estimator(self):
        """Returns base learner using its origin and the given hyperparameters

        Returns:
            est (estimator): Estimator object
        """
        estimator = self.base_learner_origin.return_estimator()
        estimator = estimator.set_params(**self.hyperparameters)
        return estimator

    def meta_features_path(self, path):
        """Returns path for meta-features

        Args:
            path (str): Absolute/local path of xcessiv folder
        """
        return os.path.join(
                path,
                app.config['XCESSIV_META_FEATURES_FOLDER'],
                str(self.id)
            ) + '.npy'

    @property
    def serialize(self):
        return dict(
            id=self.id,
            hyperparameters=self.hyperparameters,
            individual_score=self.individual_score,
            job_status=self.job_status,
            job_id=self.job_id,
            description=self.description,
            meta_features_exists=self.meta_features_exists,
            base_learner_origin_id=self.base_learner_origin_id
        )

    def delete_meta_features(self, path):
        """Deletes meta-features of base learner if it exists

        Args:
            path (str): Absolute/local path of xcessiv folder
        """
        if os.path.exists(self.meta_features_path(path)):
            os.remove(self.meta_features_path(path))

    def cleanup(self, path):
        """This function should be called before database deletion to do any pre-delete work

        Args:
            path (str, unicode): Absolute/local path of xcessiv folder
        """
        self.delete_meta_features(path)


class StackedEnsemble(Base):
    """This table contains StackedEnsembles created in the xcessiv notebook"""
    __tablename__ = 'stackedensemble'

    id = Column(Integer, primary_key=True)
    base_learners = relationship(
        "BaseLearner",
        secondary=association_table,
        back_populates='stacked_ensembles'
    )
    base_learner_origin_id = Column(Integer, ForeignKey('baselearnerorigin.id'))
    base_learner_origin = relationship('BaseLearnerOrigin', back_populates='stacked_ensembles')
    secondary_learner_hyperparameters = Column(JsonEncodedDict)
    individual_score = Column(JsonEncodedDict)
    append_original = Column(Boolean)
    job_status = Column(Text)
    job_id = Column(Text)
    description = Column(JsonEncodedDict)

    def __init__(self, secondary_learner_hyperparameters, base_learners,
                 base_learner_origin, append_original, job_status):
        self.base_learner_origin = base_learner_origin
        self.secondary_learner_hyperparameters = secondary_learner_hyperparameters
        self.base_learners = base_learners
        self.append_original = append_original
        self.individual_score = dict()
        self.job_status = job_status
        self.job_id = None
        self.description = dict()

    def return_secondary_learner(self):
        """Returns secondary learner using its origin and the given hyperparameters

        Returns:
            est (estimator): Estimator object
        """
        estimator = self.base_learner_origin.return_estimator()
        estimator = estimator.set_params(**self.secondary_learner_hyperparameters)
        return estimator

    @property
    def serialize(self):
        return dict(
            id=self.id,
            secondary_learner_hyperparameters=self.secondary_learner_hyperparameters,
            individual_score=self.individual_score,
            job_status=self.job_status,
            job_id=self.job_id,
            description=self.description,
            base_learner_origin_id=self.base_learner_origin_id,
            base_learner_ids=map(lambda x: x.id, self.base_learners),
            number_of_base_learners=len(self.base_learners),
            append_original=self.append_original
        )



"""This module contains preset settings for defining cross-validation iterators"""
from __future__ import absolute_import, print_function, division, unicode_literals


__all__ = [
    'k_fold',
    'stratified_k_fold',
    'shuffle_split',
    'stratified_shuffle_split'
]

k_fold = {
    'name': 'K-fold Cross Validation',
    'source':
    """from sklearn.model_selection import KFold

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    K times into different train-test splits.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 5
    SHUFFLE = True

    return KFold(n_splits=N_SPLITS, random_state=RANDOM_STATE, shuffle=SHUFFLE).split(X, y)
"""
}

stratified_k_fold = {
    'name': 'Stratified K-fold Cross Validation',
    'source':
    """from sklearn.model_selection import StratifiedKFold

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    K times into different stratified train-test splits.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 5
    SHUFFLE = True

    return StratifiedKFold(n_splits=N_SPLITS, random_state=RANDOM_STATE, shuffle=SHUFFLE).split(X, y)
"""
}

shuffle_split = {
    'name': 'Shuffle Split',
    'source':
    """from sklearn.model_selection import ShuffleSplit

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    randomly into different train-test splits. For applications where you
    only need a single train-test split (large datasets), you can do this
    by setting `n_splits` to 1.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 1
    TEST_SIZE = 0.25
    TRAIN_SIZE = None

    return ShuffleSplit(n_splits=N_SPLITS, random_state=RANDOM_STATE,
                        test_size=TEST_SIZE, train_size=TRAIN_SIZE).split(X, y)
"""
}

stratified_shuffle_split = {
    'name': 'Stratified Shuffle Split',
    'source':
    """from sklearn.model_selection import StratifiedShuffleSplit

def return_splits_iterable(X, y):
    \"\"\"This function returns an iterable that splits the given dataset
    randomly into different stratified train-test splits. For applications where you
    only need a single train-test split (large datasets), you can do this
    by setting `n_splits` to 1.
    \"\"\"
    RANDOM_STATE = 8
    N_SPLITS = 1
    TEST_SIZE = 0.25
    TRAIN_SIZE = None

    return StratifiedShuffleSplit(n_splits=N_SPLITS, random_state=RANDOM_STATE,
                                  test_size=TEST_SIZE, train_size=TRAIN_SIZE).split(X, y)
"""
}



from rq import Connection, Worker
from redis import Redis


def runworker(app):

    REDIS_HOST = app.config['REDIS_HOST']
    REDIS_PORT = app.config['REDIS_PORT']
    REDIS_DB = app.config['REDIS_DB']
    QUEUES = app.config['QUEUES']

    redis_conn = Connection(Redis(REDIS_HOST,
                                  REDIS_PORT,
                                  REDIS_DB))
    with redis_conn:
        w = Worker(QUEUES)
        w.work()



"""This module contains preset settings for defining metric generators"""
from __future__ import absolute_import, print_function, division, unicode_literals


__all__ = [
    'accuracy_from_scores',
    'accuracy_from_preds',
    'recall_from_scores',
    'recall_from_preds',
    'precision_from_scores',
    'precision_from_preds',
    'f1_score_from_scores',
    'f1_score_from_preds',
    'mse',
    'roc_auc_score_from_scores'
]


accuracy_from_scores = {
    'name': 'Accuracy',
    'source':
    """from sklearn.metrics import accuracy_score
import numpy as np

def metric_generator(y_true, y_probas):
    \"\"\"This function computes the accuracy given the true labels array (y_true)
    and the scores/probabilities array (y_probas) with shape (num_samples, num_classes).
    For the function to work correctly, the columns of the probabilities array must
    correspond to a sorted set of the unique values present in y_true.
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) != y_probas.shape[1]:
        raise ValueError('The shape of y_probas does not correspond to the number of unique values in y_true')
    argmax = np.argmax(y_probas, axis=1)
    y_preds = classes_[argmax]
    return accuracy_score(y_true, y_preds)
""",
    'selection_name': 'Accuracy from Scores/Probabilities'
}

accuracy_from_preds = {
    'name': 'Accuracy',
    'source':
    """from sklearn.metrics import accuracy_score

def metric_generator(y_true, y_preds):
    \"\"\"This function computes the accuracy given the true labels array (y_true)
    and the predicted labels array (y_preds).
    \"\"\"
    return accuracy_score(y_true, y_preds)
""",
    'selection_name': 'Accuracy from Predictions'
}

recall_from_scores = {
    'name': 'Recall',
    'source':
    """from sklearn.metrics import recall_score
import numpy as np

def metric_generator(y_true, y_probas):
    \"\"\"This function computes the recall given the true labels array (y_true)
    and the scores/probabilities array (y_probas) with shape (num_samples, num_classes).
    For the function to work correctly, the columns of the probabilities array must
    correspond to a sorted set of the unique values present in y_true. If there are more than
    two classes, micro-averaging is used by default.
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) != np.array(y_probas).shape[1]:
        raise ValueError('The shape of y_probas does not correspond to the number of unique values in y_true')
    argmax = np.argmax(y_probas, axis=1)
    y_preds = classes_[argmax]
    if np.array(y_probas).shape[1] > 2:
        score = recall_score(y_true, y_preds, average='micro')
    else:
        score = recall_score(y_true, y_preds)
    return score
""",
    'selection_name': 'Recall from Scores/Probabilities'
}


recall_from_preds = {
    'name': 'Recall',
    'source':
    """from sklearn.metrics import recall_score
import numpy as np

def metric_generator(y_true, y_preds):
    \"\"\"This function computes the recall given the true labels array (y_true)
    and the predicted labels array (y_preds).
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) > 2:
        score = recall_score(y_true, y_preds, average='micro')
    else:
        score = recall_score(y_true, y_preds)
    return score
""",
    'selection_name': 'Recall from Predictions'
}

precision_from_scores = {
    'name': 'Precision',
    'source':
    """from sklearn.metrics import precision_score
import numpy as np

def metric_generator(y_true, y_probas):
    \"\"\"This function computes the precision given the true labels array (y_true)
    and the scores/probabilities array (y_probas) with shape (num_samples, num_classes).
    For the function to work correctly, the columns of the probabilities array must
    correspond to a sorted set of the unique values present in y_true. If there are more than
    two classes, micro-averaging is used by default.
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) != np.array(y_probas).shape[1]:
        raise ValueError('The shape of y_probas does not correspond to the number of unique values in y_true')
    argmax = np.argmax(y_probas, axis=1)
    y_preds = classes_[argmax]
    if np.array(y_probas).shape[1] > 2:
        score = precision_score(y_true, y_preds, average='micro')
    else:
        score = precision_score(y_true, y_preds)
    return score
""",
    'selection_name': 'Precision from Scores/Probabilities'
}


precision_from_preds = {
    'name': 'Precision',
    'source':
    """from sklearn.metrics import precision_score
import numpy as np

def metric_generator(y_true, y_preds):
    \"\"\"This function computes the precision given the true labels array (y_true)
    and the predicted labels array (y_preds).
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) > 2:
        score = precision_score(y_true, y_preds, average='micro')
    else:
        score = precision_score(y_true, y_preds)
    return score
""",
    'selection_name': 'Precision from Predictions'
}

f1_score_from_scores = {
    'name': 'F1 Score',
    'source':
    """from sklearn.metrics import f1_score
import numpy as np

def metric_generator(y_true, y_probas):
    \"\"\"This function computes the F1 score given the true labels array (y_true)
    and the scores/probabilities array (y_probas) with shape (num_samples, num_classes).
    For the function to work correctly, the columns of the probabilities array must
    correspond to a sorted set of the unique values present in y_true. If there are more than
    two classes, micro-averaging is used by default.
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) != np.array(y_probas).shape[1]:
        raise ValueError('The shape of y_probas does not correspond to the number of unique values in y_true')
    argmax = np.argmax(y_probas, axis=1)
    y_preds = classes_[argmax]
    if np.array(y_probas).shape[1] > 2:
        score = f1_score(y_true, y_preds, average='micro')
    else:
        score = f1_score(y_true, y_preds)
    return score
""",
    'selection_name': 'F1 Score from Scores/Probabilities'
}


f1_score_from_preds = {
    'name': 'F1 Score',
    'source':
    """from sklearn.metrics import f1_score
import numpy as np

def metric_generator(y_true, y_preds):
    \"\"\"This function computes the F1 score given the true labels array (y_true)
    and the predicted labels array (y_preds).
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) > 2:
        score = f1_score(y_true, y_preds, average='micro')
    else:
        score = f1_score(y_true, y_preds)
    return score
""",
    'selection_name': 'F1 Score from Predictions'
}

mse = {
    'name': 'Mean Squared Error',
    'source':
    """from sklearn.metrics import mean_squared_error

metric_generator = mean_squared_error
""",
    'selection_name': 'Mean Squared Error'
}


roc_auc_score_from_scores = {
    'name': 'ROC AUC Score',
    'source':
    """from sklearn.metrics import roc_auc_score
from sklearn.preprocessing import label_binarize
import numpy as np

def metric_generator(y_true, y_probas):
    \"\"\"This function computes the Area under Curve of the
    Receiver Operating Characteristic given the true labels array (y_true)
    and the scores/probabilities array (y_probas). In a non-binary classification
    task, this will calculate a weighted mean of the AUC for each class. This
    behavior can be changed by passing a different parameter to the
    `average` argument.
    \"\"\"
    classes_ = np.unique(y_true)
    if len(classes_) != np.array(y_probas).shape[1]:
        raise ValueError('The shape of y_probas does not correspond to the number of unique values in y_true')
    binarized = label_binarize(y_true, classes_)
    if len(classes_) == 2:
        binarized = binarized.ravel()
    return roc_auc_score(binarized, y_probas, average='weighted')
""",
    'selection_name': 'ROC AUC Score from Scores/Probabilities'
}



from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import json
import os
from xcessiv import app, functions, models, constants


class TestCreateNewEnsemble(unittest.TestCase):
    def setUp(self):
        self.app = app.test_client()
        self.test_location = 'test_folder'

    def tearDown(self):
        if os.path.exists(os.path.join(self.test_location,
                                       app.config['XCESSIV_NOTEBOOK_NAME'])):
            os.remove(os.path.join(self.test_location,
                                   app.config['XCESSIV_NOTEBOOK_NAME']))
            os.rmdir(self.test_location)

    def test_creation(self):
        rv = self.app.post(
            '/ensemble/',
            data=json.dumps(
                {
                    'ensemble_name': self.test_location
                }
            ),
            content_type='application/json'
        )
        assert rv.status_code == 200
        assert os.path.exists(os.path.join(self.test_location,
                                           app.config['XCESSIV_NOTEBOOK_NAME']))
        with functions.DBContextManager(self.test_location) as session:
            extraction = session.query(models.Extraction).all()
            assert len(extraction) == 1
            assert extraction[0].main_dataset == constants.DEFAULT_EXTRACTION_MAIN_DATASET
            assert extraction[0].test_dataset == constants.DEFAULT_EXTRACTION_TEST_DATASET
            assert extraction[0].meta_feature_generation == \
                constants.DEFAULT_EXTRACTION_META_FEATURE_GENERATION

    def test_duplicate(self):
        rv = self.app.post(
            '/ensemble/',
            data=json.dumps(
                {
                    'ensemble_name': self.test_location
                }
            ),
            content_type='application/json'
        )

        rv = self.app.post(
            '/ensemble/',
            data=json.dumps(
                {
                    'ensemble_name': self.test_location
                }
            ),
            content_type='application/json'
        )

        assert rv.status_code == 400



"""This module contains preset source codes for base learners"""
from __future__ import absolute_import, print_function, division, unicode_literals
___all__ = [
    'sklearn_random_forest_source'
]

sklearn_random_forest_source = \
    """from sklearn.ensemble import RandomForestClassifier

base_learner = RandomForestClassifier(random_state=8)
"""

sklearn_extra_trees_source = \
    """from sklearn.ensemble import ExtraTreesClassifier

base_learner = ExtraTreesClassifier(random_state=8)
"""

sklearn_logistic_regression_source = \
    """from sklearn.linear_model import LogisticRegression

base_learner = LogisticRegression()
"""

sklearn_knn_classifier_source = \
    """from sklearn.neighbors import KNeighborsClassifier

base_learner = KNeighborsClassifier()
"""

sklearn_svm_classifier_source = \
    """from sklearn.svm import SVC

base_learner = SVC(random_state=8)
"""

sklearn_gaussian_nb_source = \
    """from sklearn.naive_bayes import GaussianNB

base_learner = GaussianNB()
"""

sklearn_adaboost_classifier_source = \
    """from sklearn.ensemble import AdaBoostClassifier

base_learner = AdaBoostClassifier(random_state=8)
"""

xgboost_classifier_source = \
    """from xgboost import XGBClassifier

base_learner = XGBClassifier(seed=8)
"""

xgboost_regressor_source = \
    """from xgboost import XGBRegressor

base_learner = XGBRegressor(seed=8)
"""

sklearn_linear_regression_source = \
    """from sklearn.linear_model import LinearRegression

base_learner = LinearRegression()
"""

sklearn_adaboost_regressor_source = \
    """from sklearn.ensemble import AdaBoostRegressor

base_learner = AdaBoostRegressor(random_state=8)
"""



from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
from xcessiv import models
from sklearn.ensemble import RandomForestClassifier


class TestReturnTrainDataFromJSON(unittest.TestCase):
    def setUp(self):
        self.extraction = models.Extraction()
        self.extraction.main_dataset['source'] = ''.join([
            "from sklearn.datasets import load_digits\n",
            "\n",
            "\n",
            "def extract_main_dataset():\n",
            "    X, y = load_digits(return_X_y=True)\n",
            "    return X, y"
        ])

    def test_main_is_train(self):
        X, y = self.extraction.return_train_dataset()
        assert X.shape == (1797, 64)
        assert y.shape == (1797,)

    def test_split_main_for_test(self):
        self.extraction.test_dataset['method'] = 'split_from_main'
        self.extraction.test_dataset['split_ratio'] = 0.1
        self.extraction.test_dataset['split_seed'] = 8
        X, y = self.extraction.return_train_dataset()
        assert X.shape == (1617, 64)
        assert y.shape == (1617,)


class TestReturnTestDataFromJSON(unittest.TestCase):
    def setUp(self):
        self.extraction = models.Extraction()
        self.extraction.main_dataset['source'] = ''.join([
            "from sklearn.datasets import load_digits\n",
            "\n",
            "\n",
            "def extract_main_dataset():\n",
            "    X, y = load_digits(return_X_y=True)\n",
            "    return X, y"
        ])
        self.extraction.test_dataset['method'] = 'split_from_main'
        self.extraction.test_dataset['split_ratio'] = 0.1
        self.extraction.test_dataset['split_seed'] = 8

    def test_split_main_for_test(self):
        X, y = self.extraction.return_test_dataset()
        assert X.shape == (180, 64)
        assert y.shape == (180,)

    def test_test_dataset_from_source(self):
        self.extraction.test_dataset["method"] = "source"
        self.extraction.test_dataset["source"] = ''.join([
            "from sklearn.datasets import load_digits\n",
            "def extract_test_dataset():\n",
            "    X, y = load_digits(return_X_y=True)\n",
            "    return X, y"
        ])
        X, y = self.extraction.return_test_dataset()
        assert X.shape == (1797, 64)
        assert y.shape == (1797,)


class TestReturnEstimator(unittest.TestCase):
    def setUp(self):
        self.base_learner_origin = models.BaseLearnerOrigin(
            source=''.join([
                "from sklearn.ensemble import RandomForestClassifier\n",
                "base_learner = RandomForestClassifier(random_state=8)"
            ])
        )

    def test_return_estimator_from_json(self):
        est = self.base_learner_origin.return_estimator()
        assert isinstance(est, RandomForestClassifier)



from __future__ import absolute_import, print_function, division, unicode_literals
import os
import sys
import argparse
from multiprocessing import Process, Pipe
from redis import Redis
from xcessiv.server import launch
from xcessiv.scripts.runworker import runworker
from xcessiv import app
from six import iteritems


def wrap(task, pipe):
    def wrapper(*args, **kwargs):
        sys.stdout = pipe
        task(*args, **kwargs)
    return wrapper


def main():
    parser = argparse.ArgumentParser(description='Launch Xcessiv server and workers')
    parser.add_argument('-w', '--worker', help='Define number of workers', type=int)
    parser.add_argument('-p', '--port', help='Port number to be used by web server',
                        type=int)
    parser.add_argument('-H', '--host', help='Redis host')
    parser.add_argument('-P', '--redisport', help='Redis port', type=int)
    parser.add_argument('-D', '--redisdb', help='Redis database number', type=int)
    args = parser.parse_args()

    # Check if Windows
    if os.name == 'nt':
        raise OSError('Xcessiv has detected that you are using Windows. '
                      'Unfortunately, Xcessiv does not currently support Windows. '
                      'The accepted workaround for this is to use Docker to run '
                      'Xcessiv instead. Please check the Xcessiv documentation for '
                      'more details.')

    # Overwrite configuration from configuration file
    default_config_path = os.path.join(os.path.expanduser('~'), '.xcessiv/config.py')
    if os.path.exists(default_config_path):
        print('Config file found at ' + default_config_path)
        app.config.from_pyfile(default_config_path)

    # Overwrite configuration from command line arguments
    cli_config = {
        'REDIS_HOST': args.host,
        'REDIS_PORT': args.redisport,
        'REDIS_DB': args.redisdb,
        'XCESSIV_PORT': args.port,
        'NUM_WORKERS': args.worker
    }
    cli_config = dict((key, value) for key, value in iteritems(cli_config) if value is not None)
    app.config.update(**cli_config)

    redis_conn = (Redis(app.config['REDIS_HOST'],
                        app.config['REDIS_PORT'],
                        app.config['REDIS_DB']))
    redis_conn.get(None)  # will throw exception if Redis is unavailable

    cwd = os.getcwd()
    print(cwd)

    processes = []
    try:
        conn1, conn2 = Pipe(duplex=False)
        server_proc = Process(target=wrap(launch, conn2), args=(app,))
        server_proc.start()

        for i in range(app.config['NUM_WORKERS']):
            p = Process(target=wrap(runworker, conn2), args=(app,))
            processes.append(p)
            p.start()

        server_proc.join()
    finally:
        for proc in processes:
            proc.terminate()
            proc.join()
        server_proc.terminate()
        server_proc.join()


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, division, unicode_literals
import imp
import sys
import os
import hashlib
import json
import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from six import exec_, iteritems
from sklearn import datasets
from sklearn import model_selection
from xcessiv import app, exceptions


def hash_file(path, block_size=65536):
    """Returns SHA256 checksum of a file

    Args:
        path (string): Absolute file path of file to hash

        block_size (int, optional): Number of bytes to read per block
    """
    sha256 = hashlib.sha256()
    with open(path, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()


def hash_string(string):
    """Hashes an input string using SHA256"""
    return hashlib.sha256(string).hexdigest()


def import_object_from_path(path, object):
    """Used to import an object from an absolute path.

    This function takes an absolute path and imports it as a Python module.
    It then returns the object with name `object` from the imported module.

    Args:
        path (string): Absolute file path of .py file to import

        object (string): Name of object to extract from imported module
    """
    with open(path) as f:
        return import_object_from_string_code(f.read(), object)


def import_object_from_string_code(code, object):
    """Used to import an object from arbitrary passed code.

    Passed in code is treated as a module and is imported and added
    to `sys.modules` with its SHA256 hash as key.

    Args:
        code (string): Python code to import as module

        object (string): Name of object to extract from imported module
    """
    sha256 = hashlib.sha256(code.encode('UTF-8')).hexdigest()
    module = imp.new_module(sha256)
    try:
        exec_(code, module.__dict__)
    except Exception as e:
        raise exceptions.UserError('User code exception', exception_message=str(e))
    sys.modules[sha256] = module
    try:
        return getattr(module, object)
    except AttributeError:
        raise exceptions.UserError("{} not found in code".format(object))


def verify_dataset(X, y):
    """Verifies if a dataset is valid for use i.e. scikit-learn format

    Used to verify a dataset by returning shape and basic statistics of
    returned data. This will also provide quick and dirty check on
    capability of host machine to process the data.

    Args:
        X (array-like): Features array

        y (array-like): Label array

    Returns:
        X_shape (2-tuple of int): Shape of X returned

        y_shape (1-tuple of int): Shape of y returned

    Raises:
        AssertionError: `X_shape` must be of length 2 and `y_shape` must be of
            length 1. `X` must have the same number of elements as `y`
            i.e. X_shape[0] == y_shape[0]. If any of these conditions are not met,
            an AssertionError is raised.
    """
    X_shape, y_shape = np.array(X).shape, np.array(y).shape
    if len(X_shape) != 2:
        raise exceptions.UserError("X must be 2-dimensional array")
    if len(y_shape) != 1:
        raise exceptions.UserError("y must be 1-dimensional array")
    if X_shape[0] != y_shape[0]:
        raise exceptions.UserError("X must have same number of elements as y")
    return dict(
        features_shape=X_shape,
        labels_shape=y_shape
    )


def is_valid_json(x):
    """Returns true if x can be JSON serialized

    Args:
        x: Object to test
    """
    try:
        json.dumps(x)
        return True
    except TypeError:
        return False


def make_serializable(json):
    """This function ensures that the dictionary is JSON serializable. If not,
    keys with non-serializable values are removed from the return value.

    Args:
        json (dict): Dictionary to convert to serializable

    Returns:
        new_dict (dict): New dictionary with non JSON serializable values removed
    """
    new_dict = dict()
    for key, value in iteritems(json):
        if is_valid_json(value):
            new_dict[key] = value

    return new_dict


def get_sample_dataset(dataset_properties):
    """Returns sample dataset

    Args:
        dataset_properties (dict): Dictionary corresponding to the properties of the dataset
            used to verify the estimator and metric generators.

    Returns:
        X (array-like): Features array

        y (array-like): Labels array

        splits (iterator): This is an iterator that returns train test splits for
            cross-validation purposes on ``X`` and ``y``.
    """
    kwargs = dataset_properties.copy()
    data_type = kwargs.pop('type')
    if data_type == 'multiclass':
        try:
            X, y = datasets.make_classification(random_state=8, **kwargs)
            splits = model_selection.StratifiedKFold(n_splits=2, random_state=8).split(X, y)
        except Exception as e:
            raise exceptions.UserError(repr(e))
    elif data_type == 'iris':
        X, y = datasets.load_iris(return_X_y=True)
        splits = model_selection.StratifiedKFold(n_splits=2, random_state=8).split(X, y)
    elif data_type == 'mnist':
        X, y = datasets.load_digits(return_X_y=True)
        splits = model_selection.StratifiedKFold(n_splits=2, random_state=8).split(X, y)
    elif data_type == 'breast_cancer':
        X, y = datasets.load_breast_cancer(return_X_y=True)
        splits = model_selection.StratifiedKFold(n_splits=2, random_state=8).split(X, y)
    elif data_type == 'boston':
        X, y = datasets.load_boston(return_X_y=True)
        splits = model_selection.KFold(n_splits=2, random_state=8).split(X)
    elif data_type == 'diabetes':
        X, y = datasets.load_diabetes(return_X_y=True)
        splits = model_selection.KFold(n_splits=2, random_state=8).split(X)
    else:
        raise exceptions.UserError('Unknown dataset type {}'.format(dataset_properties['type']))
    return X, y, splits


def verify_estimator_class(est, meta_feature_generator, metric_generators, dataset_properties):
    """Verify if estimator object is valid for use i.e. scikit-learn format

    Verifies if an estimator is fit for use by testing for existence of methods
    such as `get_params` and `set_params`. Must also be able to properly fit on
    and predict a sample iris dataset.

    Args:
        est: Estimator object with `fit`, `predict`/`predict_proba`,
            `get_params`, and `set_params` methods.

        meta_feature_generator (str, unicode): Name of the method used by the estimator
            to generate meta-features on a set of data.

        metric_generators (dict): Dictionary of key value pairs where the key
            signifies the name of the metric calculated and the value is a list
            of strings, when concatenated, form Python code containing the
            function used to calculate the metric from true values and the
            meta-features generated.

        dataset_properties (dict): Dictionary corresponding to the properties of the dataset
            used to verify the estimator and metric generators.

    Returns:
        performance_dict (mapping): Mapping from performance metric
            name to performance metric value e.g. "Accuracy": 0.963

        hyperparameters (mapping): Mapping from the estimator's hyperparameters to
            their default values e.g. "n_estimators": 10
    """
    X, y, splits = get_sample_dataset(dataset_properties)

    if not hasattr(est, "get_params"):
        raise exceptions.UserError('Estimator does not have get_params method')
    if not hasattr(est, "set_params"):
        raise exceptions.UserError('Estimator does not have set_params method')
    if not hasattr(est, meta_feature_generator):
        raise exceptions.UserError('Estimator does not have meta-feature generator'
                                   ' {}'.format(meta_feature_generator))

    performance_dict = dict()

    true_labels = []
    preds = []

    try:
        for train_index, test_index in splits:
            X_train, X_test = X[train_index], X[test_index]
            y_train, y_test = y[train_index], y[test_index]
            est.fit(X_train, y_train)
            true_labels.append(y_test)
            preds.append(getattr(est, meta_feature_generator)(X_test))
        true_labels = np.concatenate(true_labels)
        preds = np.concatenate(preds, axis=0)
    except Exception as e:
        raise exceptions.UserError(repr(e))

    if preds.shape[0] != true_labels.shape[0]:
        raise exceptions.UserError('Estimator\'s meta-feature generator '
                                   'does not produce valid shape')

    for key in metric_generators:
        metric_generator = import_object_from_string_code(
            metric_generators[key],
            'metric_generator'
        )
        try:
            performance_dict[key] = metric_generator(true_labels, preds)
        except Exception as e:
            raise exceptions.UserError(repr(e))

    return performance_dict, make_serializable(est.get_params())


def get_path_from_query_string(req):
    """Gets path from query string

    Args:
        req (flask.request): Request object from Flask

    Returns:
        path (str): Value of "path" parameter from query string

    Raises:
        exceptions.UserError: If "path" is not found in query string
    """
    if req.args.get('path') is None:
        raise exceptions.UserError('Path not found in query string')
    return req.args.get('path')


class DBContextManager():
    """Use this context manager to automatically start and close a database session

    Examples:
        >>> with DBContextManager('ProjectFolder') as session:
        >>>     # Do stuff with session
    """
    def __init__(self, path):
        """Initialize context manager

        Args:
            path (str, unicode): Path to project folder
        """
        self.path = os.path.join(path, app.config['XCESSIV_NOTEBOOK_NAME'])

    def __enter__(self):
        if not os.path.exists(self.path):
            raise exceptions.UserError('{} does not exist'.format(self.path))
        sqlite_url = 'sqlite:///{}'.format(self.path)
        engine = create_engine(sqlite_url)

        self.session = Session(bind=engine)

        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, 'session'):
            if exc_type is not None:
                self.session.rollback()
            self.session.close()
        return False  # re-raise any exception



from __future__ import absolute_import, print_function, division, unicode_literals
import unittest
import os
import numpy as np
from xcessiv import functions, exceptions
from sklearn.datasets import load_digits
from sklearn.ensemble import RandomForestClassifier
from sklearn.decomposition import PCA
from sklearn.pipeline import Pipeline
import pickle


filepath = os.path.join(os.path.dirname(__file__),
                        'extractmaindataset.py')


class TestHashFile(unittest.TestCase):
    def test_hash_file(self):
        assert functions.hash_file(filepath) == "8f562f857f8b13d7e2b1f2ac59d2fc" \
                                                "7603ba47db1cacef1d16ed7730102af5a7"

        assert functions.hash_file(filepath) == functions.hash_file(filepath, 2)


class TestImportObjectFromPath(unittest.TestCase):
    def test_import_object_from_path(self):
        returned_object = functions.import_object_from_path(filepath,
                                                            "extract_main_dataset")
        assert callable(returned_object)

        pickle.loads(pickle.dumps(returned_object))  # make sure pickle works


class TestImportObjectFromStringCode(unittest.TestCase):
    def test_import_object_from_string_code(self):
        with open(filepath) as f:
            returned_object = functions.\
                import_object_from_string_code(f.read(), "extract_main_dataset")

        assert callable(returned_object)

        pickle.loads(pickle.dumps(returned_object))  # make sure pickle works


class TestVerifyDataset(unittest.TestCase):
    def test_correct_dataset(self):
        X, y = load_digits(return_X_y=True)
        verification_dict = functions.verify_dataset(X, y)
        assert verification_dict['features_shape'] == (1797,64)
        assert verification_dict['labels_shape'] == (1797,)

    def test_invalid_assertions(self):
        self.assertRaises(exceptions.UserError,
                          functions.verify_dataset,
                          [[1, 2, 2], [2, 3, 5]],
                          [1, 2, 3])

        self.assertRaises(exceptions.UserError,
                          functions.verify_dataset,
                          [[1, 2, 2], [2, 3, 5]],
                          [[1, 2, 3]])

        self.assertRaises(exceptions.UserError,
                          functions.verify_dataset,
                          [[[1, 2, 2]], [[2, 3, 5]]],
                          [1, 2, 3])


class TestIsValidJSON(unittest.TestCase):
    def test_is_valid_json(self):
        assert functions.is_valid_json({'x': ['i am serializable', 0.1]})
        assert not functions.is_valid_json({'x': RandomForestClassifier()})


class TestMakeSerializable(unittest.TestCase):
    def test_make_serializable(self):
        assert functions.is_valid_json({'x': ['i am serializable', 0.1]})
        assert not functions.is_valid_json({'x': RandomForestClassifier()})
        assert functions.make_serializable(
            {
                'x': ['i am serializable', 0.1],
                'y': RandomForestClassifier()
            }
        ) == {'x': ['i am serializable', 0.1]}


class GetSampleDataset(unittest.TestCase):
    def setUp(self):
        self.dataset_properties = {
            'type': 'multiclass',
        }

    def test_classification_dataset(self):
        X, y, split = functions.get_sample_dataset(self.dataset_properties)
        assert X.shape == (100, 20)
        assert y.shape == (100,)
        assert len(np.unique(y)) == 2

        self.dataset_properties['n_classes'] = 4
        self.dataset_properties['n_informative'] = 18
        X, y, split = functions.get_sample_dataset(self.dataset_properties)
        assert X.shape == (100, 20)
        assert y.shape == (100,)
        assert len(np.unique(y)) == 4

        self.dataset_properties['n_features'] = 100
        X, y, split = functions.get_sample_dataset(self.dataset_properties)
        assert X.shape == (100, 100)
        assert y.shape == (100,)
        assert len(np.unique(y)) == 4

        self.dataset_properties['n_samples'] = 24
        X, y, split = functions.get_sample_dataset(self.dataset_properties)
        assert X.shape == (24, 100)
        assert y.shape == (24,)
        assert len(np.unique(y)) == 4

    def test_iris_dataset(self):
        X, y, split = functions.get_sample_dataset({'type': 'iris'})
        assert X.shape == (150, 4)
        assert y.shape == (150,)

    def test_mnist_dataset(self):
        X, y, split = functions.get_sample_dataset({'type': 'mnist'})
        assert X.shape == (1797, 64)
        assert y.shape == (1797,)

    def test_breast_cancer_dataset(self):
        X, y, split = functions.get_sample_dataset({'type': 'breast_cancer'})
        assert X.shape == (569, 30)
        assert y.shape == (569,)

    def test_boston_housing(self):
        X, y, split = functions.get_sample_dataset({'type': 'boston'})
        assert X.shape == (506, 13)
        assert y.shape == (506,)

    def test_diabetes(self):
        X, y, split = functions.get_sample_dataset({'type': 'diabetes'})
        assert X.shape == (442, 10)
        assert y.shape == (442,)


class TestVerifyEstimatorClass(unittest.TestCase):
    def setUp(self):
        self.source = ''.join([
            "from sklearn.metrics import accuracy_score\n",
            "import numpy as np\n",
            "def metric_generator(y_true, y_probas):\n",
            "    argmax = np.argmax(y_probas, axis=1)\n",
            "    return accuracy_score(y_true, argmax)"
        ])
        self.wrong_source = "metric_generator = ''"
        self.dataset_properties = {
            'type': 'multiclass',
        }

    def test_verify_estimator_class(self):
        np.random.seed(8)
        performance_dict, hyperparameters = functions.verify_estimator_class(
            RandomForestClassifier(),
            'predict_proba',
            dict(Accuracy=self.source),
            self.dataset_properties
        )
        assert round(performance_dict['Accuracy'], 3) == 0.8
        assert hyperparameters == {
            'warm_start': False,
            'oob_score': False,
            'n_jobs': 1,
            'verbose': 0,
            'max_leaf_nodes': None,
            'bootstrap': True,
            'min_samples_leaf': 1,
            'n_estimators': 10,
            'min_samples_split': 2,
            'min_weight_fraction_leaf': 0.0,
            'criterion': 'gini',
            'random_state': None,
            'min_impurity_split': 1e-07,
            'max_features': 'auto',
            'max_depth': None,
            'class_weight': None
        }

    def test_non_serializable_parameters(self):
        pipeline = Pipeline((('pca', PCA()), ('rf', RandomForestClassifier())))
        performance_dict, hyperparameters = functions.verify_estimator_class(
            pipeline,
            'predict_proba',
            dict(Accuracy=self.source),
            self.dataset_properties
        )
        assert functions.is_valid_json(hyperparameters)

    def test_assertion_of_invalid_metric_generator(self):
        np.random.seed(8)
        self.assertRaises(
            exceptions.UserError,
            functions.verify_estimator_class,
            RandomForestClassifier(),
            'predict_proba',
            dict(Accuracy=self.wrong_source),
            self.dataset_properties
        )

    def test_assertion_meta_feature_generator(self):
        np.random.seed(8)
        self.assertRaises(
            exceptions.UserError,
            functions.verify_estimator_class,
            RandomForestClassifier(),
            'decision_function',
            dict(Accuracy=self.source),
            self.dataset_properties
        )



from sklearn.ensemble import RandomForestClassifier
import joblib


class MyClassifier(RandomForestClassifier):
    def save(self, filepath):
        joblib.dump(self, filepath, 3)

    @staticmethod
    def load(filepath):
        return joblib.load(filepath)



from __future__ import absolute_import, print_function, division, unicode_literals
from flask import Flask


__version__ = '0.2.4'


app = Flask(__name__, static_url_path='/static', static_folder='ui/build/static')
app.config.from_object('xcessiv.config')


import xcessiv.views



from __future__ import absolute_import, print_function, division, unicode_literals
import os
from flask import request, jsonify, send_from_directory, g
from redis import Redis
from rq import Connection
from sqlalchemy import create_engine
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import ParameterSampler
from xcessiv import app, functions, exceptions, models, rqtasks
from xcessiv.presets import learnersetting, metricsetting, cvsetting
import six


@app.errorhandler(exceptions.UserError)
def handle_user_error(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


def get_redis_connection():
    redis_conn = getattr(g, '_redis_connection', None)
    if redis_conn is None:
        redis_conn = (Redis(app.config['REDIS_HOST'],
                            app.config['REDIS_PORT'],
                            app.config['REDIS_DB']))
        g._redis_connection = redis_conn
    return redis_conn


@app.route('/', defaults={'path': 'index.html'})
@app.route('/<path:path>')
def home(path):
    return send_from_directory(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'ui/build',
            os.path.split(path)[0]
        ),
        os.path.split(path)[1]
    )


@app.route('/folders/', methods=['GET'])
def get_current_files():
    cwd = os.getcwd()
    return jsonify([os.path.join(cwd, d) for d in os.listdir(cwd)
                    if os.path.isdir(os.path.join(cwd, d))])


@app.route('/ensemble/', methods=['POST'])
def create_new_ensemble():
    req_body = request.get_json()
    ensemble_name = req_body['ensemble_name']

    if os.path.exists(ensemble_name):
        return jsonify(message="File/folder already exists"), 400

    os.makedirs(ensemble_name)
    xcessiv_notebook_path = os.path.join(ensemble_name, app.config['XCESSIV_NOTEBOOK_NAME'])
    sqlite_url = 'sqlite:///{}'.format(xcessiv_notebook_path)
    engine = create_engine(sqlite_url)

    models.Base.metadata.create_all(engine)

    # Initialize
    extraction = models.Extraction()
    with functions.DBContextManager(ensemble_name) as session:
        session.add(extraction)
        session.commit()

    return jsonify(message="Xcessiv notebook created")


@app.route('/ensemble/extraction/main-dataset/', methods=['GET', 'PATCH'])
def extraction_main_dataset():
    path = functions.get_path_from_query_string(request)

    if request.method == 'GET':
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            return jsonify(extraction.main_dataset)

    if request.method == 'PATCH':
        req_body = request.get_json()
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            for key, value in six.iteritems(req_body):
                extraction.main_dataset[key] = value
            session.add(extraction)
            session.commit()
            return jsonify(extraction.main_dataset)


@app.route('/ensemble/extraction/test-dataset/', methods=['GET', 'PATCH'])
def extraction_test_dataset():
    path = functions.get_path_from_query_string(request)

    if request.method == 'GET':
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            return jsonify(extraction.test_dataset)

    if request.method == 'PATCH':
        req_body = request.get_json()
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            for key, value in six.iteritems(req_body):
                extraction.test_dataset[key] = value
            session.add(extraction)
            session.commit()
            return jsonify(extraction.test_dataset)


@app.route('/ensemble/extraction/meta-feature-generation/', methods=['GET', 'PATCH'])
def extraction_meta_feature_generation():
    path = functions.get_path_from_query_string(request)

    if request.method == 'GET':
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            return jsonify(extraction.meta_feature_generation)

    if request.method == 'PATCH':
        req_body = request.get_json()
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            for key, value in six.iteritems(req_body):
                extraction.meta_feature_generation[key] = value
            session.add(extraction)
            session.commit()
            return jsonify(extraction.meta_feature_generation)


@app.route('/ensemble/extraction/stacked-ensemble-cv/', methods=['GET', 'PATCH'])
def extraction_stacked_ensemble_cv():
    path = functions.get_path_from_query_string(request)

    if request.method == 'GET':
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            return jsonify(extraction.stacked_ensemble_cv)

    if request.method == 'PATCH':
        req_body = request.get_json()
        with functions.DBContextManager(path) as session:
            extraction = session.query(models.Extraction).first()
            for key, value in six.iteritems(req_body):
                extraction.stacked_ensemble_cv[key] = value
            session.add(extraction)
            session.commit()
            return jsonify(extraction.stacked_ensemble_cv)


@app.route('/ensemble/extraction/verification/', methods=['GET', 'POST'])
def verify_full_extraction():
    """This is an experimental endpoint to simultaneously verify data
    statistics and extraction for training, test, and holdout datasets.
    With this, the other three verification methods will no longer be
    necessary.
    """
    path = functions.get_path_from_query_string(request)

    if request.method == 'POST':
        rqtasks.extraction_data_statistics(path)

    with functions.DBContextManager(path) as session:
        extraction = session.query(models.Extraction).first()
        return jsonify(extraction.data_statistics)


@app.route('/ensemble/cv-settings/', methods=['GET'])
def cv_settings():
    return jsonify([getattr(cvsetting, x) for x in cvsetting.__all__])


@app.route('/ensemble/base-learner-origins-settings/', methods=['GET'])
def base_learner_origins_settings():
    return jsonify([getattr(learnersetting, x) for x in learnersetting.__all__])


@app.route('/ensemble/metric-generators-settings/', methods=['GET'])
def metric_generators_settings():
    return jsonify([getattr(metricsetting, x) for x in metricsetting.__all__])


@app.route('/ensemble/base-learner-origins/', methods=['GET', 'POST'])
def base_learner_origins_view():
    path = functions.get_path_from_query_string(request)

    if request.method == 'GET':
        with functions.DBContextManager(path) as session:
            base_learner_origins = session.query(models.BaseLearnerOrigin).all()
            return jsonify(list(map(lambda x: x.serialize, base_learner_origins)))

    if request.method == 'POST':  # Create new base learner origin
        req_body = request.get_json()
        new_base_learner_origin = models.BaseLearnerOrigin(**req_body)

        with functions.DBContextManager(path) as session:
            session.add(new_base_learner_origin)
            session.commit()
            return jsonify(new_base_learner_origin.serialize)


@app.route('/ensemble/base-learner-origins/<int:id>/', methods=['GET', 'PATCH', 'DELETE'])
def specific_base_learner_origin(id):
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learner_origin = session.query(models.BaseLearnerOrigin).filter_by(id=id).first()
        if base_learner_origin is None:
            raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

        if request.method == 'GET':
            return jsonify(base_learner_origin.serialize)

        if request.method == 'PATCH':
            if base_learner_origin.final:
                raise exceptions.UserError('Cannot modify a final base learner origin')
            req_body = request.get_json()

            modifiable_attr = ('meta_feature_generator', 'name', 'source',
                               'metric_generators')
            for attr in modifiable_attr:
                if attr in req_body:
                    setattr(base_learner_origin, attr, req_body[attr])

            session.add(base_learner_origin)
            session.commit()
            return jsonify(base_learner_origin.serialize)

        if request.method == 'DELETE':
            base_learner_origin.cleanup(path)
            session.delete(base_learner_origin)
            session.commit()
            return jsonify(message='Deleted base learner origin')


@app.route('/ensemble/base-learner-origins/<int:id>/verify/', methods=['POST'])
def verify_base_learner_origin(id):
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learner_origin = session.query(models.BaseLearnerOrigin).filter_by(id=id).first()
        if base_learner_origin is None:
            raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

        if request.method == 'POST':
            req_body = request.get_json()
            if base_learner_origin.final:
                raise exceptions.UserError('Base learner origin {} '
                                           'is already final'.format(id))
            base_learner = base_learner_origin.return_estimator()
            validation_results, hyperparameters = functions.verify_estimator_class(
                base_learner,
                base_learner_origin.meta_feature_generator,
                base_learner_origin.metric_generators,
                req_body['dataset_properties']
            )
            base_learner_origin.validation_results = {
                'dataset': req_body['dataset_properties'],
                'metrics': validation_results
            }
            base_learner_origin.hyperparameters = hyperparameters
            session.add(base_learner_origin)
            session.commit()
            return jsonify(base_learner_origin.serialize)


@app.route('/ensemble/base-learner-origins/<int:id>/confirm/', methods=['GET'])
def confirm_base_learner_origin(id):
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learner_origin = session.query(models.BaseLearnerOrigin).filter_by(id=id).first()
        if base_learner_origin is None:
            raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

        if request.method == 'GET':
            if base_learner_origin.final:
                raise exceptions.UserError('Base learner origin {} '
                                           'is already final'.format(id))
            if not base_learner_origin.validation_results:
                raise exceptions.UserError('Base learner origin {} has not yet been '
                                           'verified on a dataset'.format(id))
            base_learner = base_learner_origin.return_estimator()
            validation_results, hyperparameters = functions.verify_estimator_class(
                base_learner,
                base_learner_origin.meta_feature_generator,
                base_learner_origin.metric_generators,
                base_learner_origin.validation_results['dataset']
            )
            base_learner_origin.validation_results = {
                'dataset': base_learner_origin.validation_results['dataset'],
                'metrics': validation_results
            }
            base_learner_origin.hyperparameters = hyperparameters
            base_learner_origin.final = True
            session.add(base_learner_origin)
            session.commit()
            return jsonify(base_learner_origin.serialize)


@app.route('/ensemble/base-learner-origins/<int:id>/create-base-learner/', methods=['POST'])
def create_base_learner(id):
    """This creates a single base learner from a base learner origin and queues it up"""
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learner_origin = session.query(models.BaseLearnerOrigin).filter_by(id=id).first()
        if base_learner_origin is None:
            raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

        if not base_learner_origin.final:
            raise exceptions.UserError('Base learner origin {} is not final'.format(id))

        req_body = request.get_json()

        # Retrieve full hyperparameters
        est = base_learner_origin.return_estimator()
        hyperparameters = functions.import_object_from_string_code(req_body['source'],
                                                                   'params')
        est.set_params(**hyperparameters)
        hyperparameters = functions.make_serializable(est.get_params())

        base_learners = session.query(models.BaseLearner).\
            filter_by(base_learner_origin_id=id,
                      hyperparameters=hyperparameters).all()
        if base_learners:
            raise exceptions.UserError('Base learner exists with given hyperparameters')

        base_learner = models.BaseLearner(hyperparameters,
                                          'queued',
                                          base_learner_origin)

        session.add(base_learner)
        session.commit()

        with Connection(get_redis_connection()):
            rqtasks.generate_meta_features.delay(path, base_learner.id)

        return jsonify(base_learner.serialize)


@app.route('/ensemble/base-learner-origins/<int:id>/search/', methods=['POST'])
def search_base_learner(id):
    """Creates a set of base learners from base learner origin using grid search
    and queues them up
    """
    path = functions.get_path_from_query_string(request)
    req_body = request.get_json()
    if req_body['method'] == 'grid':
        param_grid = functions.import_object_from_string_code(
            req_body['source'],
            'param_grid'
        )
        iterator = ParameterGrid(param_grid)
    elif req_body['method'] == 'random':
        param_distributions = functions.import_object_from_string_code(
            req_body['source'],
            'param_distributions'
        )
        iterator = ParameterSampler(param_distributions, n_iter=req_body['n_iter'])

    else:
        raise exceptions.UserError('{} not a valid search method'.format(req_body['method']))

    with functions.DBContextManager(path) as session:
        base_learner_origin = session.query(models.BaseLearnerOrigin).filter_by(id=id).first()
        if base_learner_origin is None:
            raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

        if not base_learner_origin.final:
            raise exceptions.UserError('Base learner origin {} is not final'.format(id))

        learners = []
        for params in iterator:
            est = base_learner_origin.return_estimator()
            try:
                est.set_params(**params)
            except Exception as e:
                print(repr(e))
                continue

            hyperparameters = functions.make_serializable(est.get_params())

            base_learners = session.query(models.BaseLearner).\
                filter_by(base_learner_origin_id=id,
                          hyperparameters=hyperparameters).all()
            if base_learners:  # already exists
                continue

            base_learner = models.BaseLearner(hyperparameters,
                                              'queued',
                                              base_learner_origin)

            session.add(base_learner)
            session.commit()
            with Connection(get_redis_connection()):
                rqtasks.generate_meta_features.delay(path, base_learner.id)
            learners.append(base_learner)
        return jsonify(list(map(lambda x: x.serialize, learners)))


@app.route('/ensemble/base-learners/', methods=['GET', 'DELETE'])
def get_base_learners():
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learners = session.query(models.BaseLearner).all()

        if request.method == 'GET':
            return jsonify(list(map(lambda x: x.serialize, base_learners)))

        if request.method == 'DELETE':  # Go crazy and delete everything
            for base_learner in base_learners:
                base_learner.delete_meta_features(path)
                session.delete(base_learner)
            session.commit()
            return jsonify(message='Deleted all base learners')


@app.route('/ensemble/base-learners/<int:id>/', methods=['GET', 'DELETE'])
def specific_base_learner(id):
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        base_learner = session.query(models.BaseLearner).filter_by(id=id).first()
        if base_learner is None:
            raise exceptions.UserError('Base learner {} not found'.format(id), 404)

        if request.method == 'GET':
            return jsonify(base_learner.serialize)

        if request.method == 'DELETE':
            base_learner.cleanup(path)
            session.delete(base_learner)
            session.commit()
            return jsonify(message='Deleted base learner')


@app.route('/ensemble/stacked/', methods=['GET', 'POST'])
def create_new_stacked_ensemble():
    path = functions.get_path_from_query_string(request)
    req_body = request.get_json()

    with functions.DBContextManager(path) as session:
        if request.method == 'GET':
                return jsonify(
                    list(map(lambda x: x.serialize, session.query(models.StackedEnsemble).all()))
                )

        if request.method == 'POST':
            base_learners = session.query(models.BaseLearner).\
                filter(models.BaseLearner.id.in_(req_body['base_learner_ids'])).all()
            if len(base_learners) != len(req_body['base_learner_ids']):
                raise exceptions.UserError('Not all base learners found')
            for learner in base_learners:
                if learner.job_status != 'finished':
                    raise exceptions.UserError('Not all base learners have finished')

            base_learner_origin = session.query(models.BaseLearnerOrigin).\
                filter_by(id=req_body['base_learner_origin_id']).first()
            if base_learner_origin is None:
                raise exceptions.UserError('Base learner origin {} not found'.format(id), 404)

            # Retrieve full hyperparameters
            est = base_learner_origin.return_estimator()
            params = functions.import_object_from_string_code\
                (req_body['secondary_learner_hyperparameters_source'], 'params')
            est.set_params(**params)
            hyperparameters = functions.make_serializable(est.get_params())

            stacked_ensemble = models.StackedEnsemble(
                secondary_learner_hyperparameters=hyperparameters,
                base_learners=base_learners,
                base_learner_origin=base_learner_origin,
                append_original=req_body['append_original'],
                job_status='queued'
            )

            session.add(stacked_ensemble)
            session.commit()

            with Connection(get_redis_connection()):
                rqtasks.evaluate_stacked_ensemble.delay(path, stacked_ensemble.id)

            return jsonify(stacked_ensemble.serialize)


@app.route('/ensemble/stacked/<int:id>/', methods=['GET', 'DELETE'])
def specific_stacked_ensemble(id):
    path = functions.get_path_from_query_string(request)

    with functions.DBContextManager(path) as session:
        stacked_ensemble = session.query(models.StackedEnsemble).filter_by(id=id).first()
        if stacked_ensemble is None:
            raise exceptions.UserError('Stacked ensemble {} not found'.format(id), 404)

        if request.method == 'GET':
            return jsonify(stacked_ensemble.serialize)

        if request.method == 'DELETE':
            session.delete(stacked_ensemble)
            session.commit()
            return jsonify(message='Deleted stacked ensemble')



from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import io
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)

long_description = read('README.md')


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    name='xcessiv',
    version='0.2.4',
    url='https://github.com/reiinakano/xcessiv',
    license='Apache License 2.0',
    author='Reiichiro Nakano',
    tests_require=['pytest'],
    install_requires=[
        'Flask>=0.11.0',
        'gevent>=1.1.0',
        'numpy>=1.12.0',
        'redis>=2.10.0',
        'rq>=0.7.0',
        'scikit-learn>=0.18.0',
        'scipy>=0.18.0',
        'six>=1.10.0',
        'SQLAlchemy>=1.1.0'
    ],
    cmdclass={'test': PyTest},
    author_email='reiichiro.s.nakano@gmail.com',
    description='A web-based application for quick and '
                'scalable construction of massive machine learning ensembles.',
    long_description=long_description,
    packages=find_packages(),
    include_package_data=True,
    platforms='any',
    test_suite='xcessiv.tests',
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: JavaScript',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
    extras_require={
        'testing': ['pytest'],
    },
    entry_points={
        'console_scripts': [
            'xcessiv = xcessiv.scripts.runapp:main'
        ]
    }
)



"""This module contains preset settings for defining base learners"""
from __future__ import absolute_import, print_function, division, unicode_literals
from xcessiv.presets import learnersource


__all__ = [
    'sklearn_random_forest',
    'sklearn_extra_trees',
    'sklearn_logistic_regression',
    'sklearn_knn_classifier',
    'sklearn_svm_classifier',
    'sklearn_gaussian_nb',
    'sklearn_adaboost_classifier',
    'xgboost_classifier',
    'xgboost_regressor',
    'sklearn_linear_regression',
    'sklearn_adaboost_regressor'
]


sklearn_random_forest = {
    'name': 'scikit-learn Random Forest',
    'source': learnersource.sklearn_random_forest_source,
    'meta_feature_generator': 'predict_proba'
}

sklearn_extra_trees = {
    'name': 'scikit-learn Extra Trees Classifier',
    'source': learnersource.sklearn_extra_trees_source,
    'meta_feature_generator': 'predict_proba'
}

sklearn_logistic_regression = {
    'name': 'scikit-learn Logistic Regression',
    'source': learnersource.sklearn_logistic_regression_source,
    'meta_feature_generator': 'predict_proba'
}

sklearn_knn_classifier = {
    'name': 'scikit-learn KNN Classifier',
    'source': learnersource.sklearn_knn_classifier_source,
    'meta_feature_generator': 'predict_proba'
}

sklearn_svm_classifier = {
    'name': 'scikit-learn SVM Classifier',
    'source': learnersource.sklearn_svm_classifier_source,
    'meta_feature_generator': 'decision_function'
}

sklearn_gaussian_nb = {
    'name': 'scikit-learn Gaussian Naive Bayes',
    'source': learnersource.sklearn_gaussian_nb_source,
    'meta_feature_generator': 'predict_proba'
}

sklearn_adaboost_classifier = {
    'name': 'scikit-learn AdaBoost Classifier',
    'source': learnersource.sklearn_adaboost_classifier_source,
    'meta_feature_generator': 'predict_proba'
}

xgboost_classifier = {
    'name': 'XGBoost Classifier',
    'source': learnersource.xgboost_classifier_source,
    'meta_feature_generator': 'predict_proba'
}

xgboost_regressor = {
    'name': 'XGBoost Regressor',
    'source': learnersource.xgboost_regressor_source,
    'meta_feature_generator': 'predict'
}

sklearn_linear_regression = {
    'name': 'scikit-learn Linear Regression',
    'source': learnersource.sklearn_linear_regression_source,
    'meta_feature_generator': 'predict'
}

sklearn_adaboost_regressor = {
    'name': 'scikit-learn AdaBoost Regressor',
    'source': learnersource.sklearn_adaboost_regressor_source,
    'meta_feature_generator': 'predict'
}



"""This module contains specific exceptions to be handled by Flask"""


class UserError(Exception):
    def __init__(self, message, status_code=400, **kwargs):
        super(UserError, self).__init__(self)
        self.message = message
        self.status_code = status_code
        self.kwargs = kwargs

    def to_dict(self):
        rv = dict(self.kwargs or ())
        rv['message'] = self.message
        return rv
